class Float
    def round_to(i)
        f = (10 ** i).to_f
        nr = self * f
        return nr.round / f
    end
end

module SkirentalFrontendApi
    class Booking
        include NoBrainer::Document
        include NoBrainer::Document::Timestamps

        Price = Struct.new(:items, :price_total, :tax_total)

        before_save :sync_with_backend

        has_many :payments, class_name:SkirentalPayment::Payment, dependent: :destroy

        field :session_id,       :type => String,                       :required => true,   :index => true
        field :shop_id,          :type => String,                       :required => true,   :index => true
        field :context,          :type => Hash,      :default => {}
        field :booking_id,       :type => String
        field :no_sync,          :type => Boolean,   :default => false
        field :changeable_until, :type => String,    :default => nil
        field :creation_time,    :type => String,    :default => nil
        field :modified_at,      :type => Time,      :default => nil
        field :ghost_time,       :type => Integer,   :default => nil
        field :is_checkin,       :type => Boolean,   :default => nil
        field :signing_url,      :type => String,    :default => nil
        field :cleared_cart,     :type => Boolean,   :default => false

        DEPOSIT_TYPE_PERCENT = 'prc'
        DEPOSIT_TYPE_FIX_PRICE_ALL = 'fxa'
        DEPOSIT_TYPE_FIX_PER_CUSTOMER = 'fxc'
        DEPOSIT_TYPE_FIX_PER_PRODUCT = 'fxp'
        DEPOSIT_TYPE_PRICE_PER_DAY = 'fdp'

        def voucher
            # TODO: Rethink whether this is good for the future. Future: If we want to update bookings or need
            # receipts, we need to store the booking transaction into the payment and have to handle multiple (old) payments
            confirmed_payment = self.payments.where(:not => { :voucher_code => nil }).order_by(:created_at => :asc).last
            if confirmed_payment.present?
                confirmed_payment.voucher_code
            else
                nil
            end
        end

        def is_paid_or_reserved
            self.payments.where(:or => [{ :is_payed => true }, { :is_reserved => true }]).any?
        end

        def transaction(tax_include = true, tax_show = true, payment_method = nil)
            fail if context.nil?

            current_locale = context[:locale] || "de"
            current_locale = current_locale.split("-").first

            shop_settings = SkirentalApi::Client.settings(shop_id, current_locale)

            currency = shop_settings[:money][:payment][:cur]

            payment_discount_factor = 1
            unless payment_method.nil?
                payment_providers = shop_settings[:money][:payment][:ipMs]
                payment_providers.each do |provider|
                    if provider[:type] == payment_method
                        payment_discount_factor = (100 - provider[:dsc].to_d) / 100
                        break
                    end
                end
            end

            start_date = Date.parse(context[:startDate]).strftime("%Y-%m-%d")
            duration = (Date.parse(context[:endDate]) - Date.parse(context[:startDate])).to_i + 1

            deposit = deposit_total(shop_settings[:money][:deposit][:value],
                                    shop_settings[:money][:deposit][:type],
                                            currency,
                                            tax_include,
                                            tax_show,
                                            current_locale,
                                            payment_discount_factor,
                                            start_date,
                                            duration)

            price = price_total(currency, tax_include, tax_show, current_locale, payment_discount_factor, start_date, duration)

            if !deposit.nil? && deposit.items.length > 0
                price = deposit
            end

            items       = price.items
            price_total = price.price_total
            tax_total   = price.tax_total

            transaction = {
                :item_list => {
                    :items => items
                },
                :amount =>  {
                    :total    => '%.2f' % ((price_total + tax_total).to_f.round_to(2)),
                    :currency => currency,
                    :details  => !tax_show ? nil : {
                        :subtotal => '%.2f' % price_total.to_f.round_to(2),
                        :tax      => '%.2f' % tax_total.to_f.round_to(2)
                    }
                },
                :description => "Wir bedanken uns fuer Ihre Buchung!" # TODO: do we need a description???
            }

            puts "CREATED TRANSACTION:"
            puts "#{transaction}"

            transaction
        end

        def already_finished
            puts "ALREADY_FINISHED?"
            puts "PAYMENTS_BLANK?: #{self.payments.blank?}"
            puts "CHANGEABLE_UNTIL: #{self.changeable_until}"
            if self.payments.blank?
                return false
            else
                puts "PAYMENT PRESENT"
                is_checkin = fetch_is_checkin()
                if is_checkin == true
                    puts "SHOP IS IN CHECKIN MODE -> WE CAN CHANGE ALL..."
                    return false
                end
                puts "SHOP IS NOT IN CHECKIN MODE"
                if self.changeable_until.blank? || self.changeable_until == "0001-01-01T00:00:00" # this value is the "zero/undefined" we get from sr-api…
                    puts "FETCHING CURRENT CHANGEABLE_UNTIL_DATE"
                    api_booking = SkirentalApi::Client.get_booking(self.booking_id)
                    api_booking_edit_until = api_booking[:editUntil]
                    puts "API_BOOKING_EDIT_UNTIL: #{api_booking_edit_until}"
                    self.changeable_until = api_booking_edit_until if api_booking_edit_until.present?
                    self.no_sync = true
                    self.save
                end
                reservation_not_changeable = self.changeable_until.to_time < Time.now
                puts "RESERVATION_NOT_CHANGEABLE: #{reservation_not_changeable}"
                self.payments.each do |payment|
                    if payment.is_payed or (payment.is_reserved and reservation_not_changeable)
                        return true
                        break
                    end
                end
                return false
            end
        end

        def sync_with_backend
            now = Time.now
            if self.modified_at.nil?
                self.modified_at = now
            end
            if self.ghost_time.nil?
                settings = SkirentalApi::Client.settings(self.shop_id)
                self.ghost_time = settings[:ghostTime].present? ? settings[:ghostTime] : 10
            end
            if self.no_sync
                self.no_sync = false
            else
                puts "SkirentalFrontendApi::Booking.sync_with_backend"
                puts "PAYMENTS_BLANK?: #{self.payments.blank?}"
                puts "CHANGEABLE_UNTIL: #{self.changeable_until}"
                puts "MODIFIED_AT: #{self.modified_at}"
                puts "GHOST_TIME: #{self.ghost_time}"
                self.cleared_cart = false
                change_finished_reservation = false # is an already finished RESERVATION allowed to be changed (editUntil >= now)?
                if self.payments.present? and self.changeable_until.present?
                    self.payments.each do |payment|
                        puts "check if #{payment} is a reservation and changeable_until is in the future"
                        if payment.is_reserved and self.changeable_until.to_time >= now
                            change_finished_reservation = true
                            break
                        end
                    end
                end
                if change_finished_reservation
                    puts "CHANGE_FINISHED_RESERVATION -> confirm_booking"
                    self.modified_at = now
                    api_booking = SkirentalApi::Client.confirm_booking(self, true)
                    puts "API_BOOKING: #{api_booking}"
                    api_booking_id = api_booking[:id]
                    self.booking_id = api_booking_id if api_booking_id.present?
                    api_booking_edit_until = api_booking[:editUntil]
                    self.changeable_until = api_booking_edit_until if api_booking_edit_until.present?
                    self.signing_url = api_booking[:signingUrl] if api_booking[:signingUrl].present?
                else
                    puts "CHANGE_UNFINISHED_RESERVATION -> sync_booking_state"
                    is_checkin = fetch_is_checkin()
                    unless is_checkin
                        valid_until_time = self.modified_at + self.ghost_time * 60
                        checkout_completed = self.context[:code].present?
                        if !checkout_completed && now > valid_until_time
                            puts "GHOST_BOOKING_EXPIRED! LOOK WHETHER WE NEED TO CLEAR THE CART"
                            self.cleared_cart = clear_cart()
                            if self.cleared_cart
                                puts "…YES, WE HAD TO CLEAR THE CART"
                            end
                        end
                    end
                    self.modified_at = now
                    api_booking = SkirentalApi::Client.sync_booking_state(self)
                    api_booking_id = api_booking[:id]
                    self.booking_id = api_booking_id if api_booking_id.present?
                    api_booking_edit_until = api_booking[:editUntil]
                    self.changeable_until = api_booking_edit_until if api_booking_edit_until.present?
                    self.signing_url = api_booking[:signingUrl] if api_booking[:signingUrl].present?
                    if self.cleared_cart
                        self.errors.add(:ghost_time, :blank, message: "EXPIRED")
                    end
                end
            end
            true
        end

        def clear_cart()
            cleared_cart = false
            self.context[:customers].each do |customer|
                if customer[:packages].present?
                    customer[:packages] = []
                    cleared_cart = true
                end
                if customer[:products].present?
                    customer[:products] = []
                    cleared_cart = true
                end
            end
            return cleared_cart
        end
        
        def fetch_is_checkin()
            if self.is_checkin.nil?
                puts "CHECK-IN FLAG NOT PRESENT YET, CALL SHOP API TO FETCH IT"
                api_shop = SkirentalApi::Client.shop(self.shop_id, "de")
                self.is_checkin = api_shop[:checkin]
                self.no_sync = true
                self.save
            end
            return self.is_checkin
        end

        private

        def fetch_price(product_id, start_date, duration, age, shop_id, current_locale, package_id)
            # 1. We expect that prices are independent to the lang.
            # 2. We store product_id's therefore we do not access prices of packages.
            price = SkirentalApi::Client.get_price_of_item(product_id, start_date, duration, age, shop_id, current_locale, package_id)
            fail if price.nil?
            return nil if price[:price].blank? || price[:price] <= 0 || price[:priceWithInsurance].blank? || price[:priceWithInsurance] <= 0

            price
        end

        def collect_customer_products(product_collector)
            context[:customers].each do |customer|
                age = customer[:customerGroup][:ageFrom]
                use_insurance = customer[:useInsurance]

                customer[:packages].each do |package|
                    # first, access the main product of a package/bundle
                    mainProductId = package[:mainProduct][:id]
                    packageId     = package[:id]
                    productCount  = package[:mainProduct][:count]

                    product_collector.call(package, mainProductId, packageId, productCount, age, use_insurance, customer)

                    # then access accessories
                    package[:accessories].each do |accessory|
                        accessoryId     = accessory[:id]
                        packageId       = package[:id]
                        accessoryCount  = accessory[:count]

                        product_collector.call(accessory, accessoryId, packageId, accessoryCount, age, use_insurance, customer)
                    end
                end

                # access single products
                customer[:products].each do |product|
                    productId     = product[:id]
                    productCount  = product[:count]

                    product_collector.call(product, productId, nil, productCount, age, use_insurance, customer)
                end
            end
        end

        def collect_customer_product_prices(start_date, duration, shop_id, current_locale, price_collector)
            collect_customer_products(lambda do |product, product_id, package_id, product_count, age, use_insurance, customer|
                price = fetch_price(product_id, start_date, duration, age, shop_id, current_locale, package_id)
                if !price.nil?
                    price_collector.call(price, product, product_id, package_id, product_count, age, use_insurance, customer)
                end
            end)
        end

        def price_total(currency, tax_include, tax_show, current_locale, payment_discount_factor, start_date, duration)
            # NOTE: tax_include and tax_show are now ignored and the included tax amound won't be listed.
            # Caclulating the tax amount separately with a possible discount factor would lead to rounding errors otherwise.
            items = []

            price_total = 0
            tax_total = 0

            collect_customer_product_prices(start_date, duration, self.shop_id, current_locale, lambda do |price, product, product_id, package_id, product_count, age, use_insurance, customer|
                product_name  = product[:name] + (age > 0 ? " [#{customer[:customerGroup][:name]}]" : "")
                description   = product[:description]

                price_value = (use_insurance ? price[:priceWithInsurance] : price[:price])
                tax_value = (use_insurance ? price[:vatValueWithInsurance] : price[:vatValue])

                # round to 2 decimals because discount_factor could lead to more decimals…
                price_value = '%.2f' % price_value.to_f.round_to(2)
                tax_value = '%.2f' % tax_value.to_f.round_to(2)

                item = {
                    :name        => product_name,
                    :description => description,
                    :quantity    => product_count,
                    :price       => price_value,
                    :tax         => nil,
                    :sku         => product_id,
                    :currency    => currency
                }

                items.push item

                price_total += item[:price].to_d * item[:quantity]
            end)

            # Calculate the discounted total price that's rounded so that the exact discount amount can be added as an item.
            price_total_discounted = (price_total * payment_discount_factor).to_f.round_to(2)
            discount_amount = price_total - price_total_discounted

            if payment_discount_factor > 0 and discount_amount > 0
                discount_percent = 1 - payment_discount_factor
                items.push({
                    :name        => "–#{(discount_percent * 100).round}%",
                    :quantity    => 1,
                    :price       => '%.2f' % -discount_amount.to_f.round_to(2),
                    :tax         => nil,
                    :sku         => nil,
                    :currency    => currency
                })
            end

            Price.new(items, price_total_discounted, 0)
        end

        def deposit_total(deposit, depositType, currency, tax_include, tax_show, current_locale, payment_discount_factor, start_date, duration)
            if deposit <= 0
                return nil
            end

            # TODO: Calculate taxes
            tax_total = 0
            deposit_value = 0
            case depositType
            when DEPOSIT_TYPE_PERCENT
                # Note that 1 is given as the discount factor such that the percentage calculation can use a price without any changes applied to it.
                price = price_total(currency, tax_include, tax_show, current_locale, 1, start_date, duration)
                deposit_value = (price.price_total / 100.0) * deposit
            when DEPOSIT_TYPE_FIX_PRICE_ALL
                deposit_value = deposit
            when DEPOSIT_TYPE_FIX_PER_CUSTOMER
                deposit_value = deposit * context[:customers].length
            when DEPOSIT_TYPE_FIX_PER_PRODUCT
                # Count the number of products.
                all_products_count = 0
                collect_customer_products(lambda do |product, product_id, package_id, product_count, age, use_insurance, customer|
                    all_products_count += product_count
                end)

                deposit_value = all_products_count * deposit
            when DEPOSIT_TYPE_PRICE_PER_DAY
                deposit_value = deposit * duration
            else
                return nil
            end

            # Calculate the possible insurance sum to add.
            insurance_sum = 0
            collect_customer_product_prices(start_date, duration, self.shop_id, current_locale, lambda do |price, product, product_id, package_id, product_count, age, use_insurance, customer|
                if use_insurance
                    insurance_sum += price[:insuranceValue] * product_count
                end
            end)

            deposit_value += insurance_sum
            deposit_value *= payment_discount_factor

            deposit_text = "Deposit"
            begin
                languageJSON = SkirentalApi::Client.i18n(current_locale, self.context[:shopLic])
                language_deposit_text = languageJSON[:data][:general][:deposit_total]

                if language_deposit_text.present?
                    deposit_text = language_deposit_text
                end
            rescue
            end

            single_deposit_item = {
                :name => deposit_text,
                :description => "",
                :quantity => 1,
                :price => "#{deposit_value}",
                :tax => nil,
                :sku => nil,
                :currency => currency
            }

            Price.new([single_deposit_item], deposit_value, tax_total)
        end
    end
end
